==> bfilter-0.10.3.tar.gz: bfilter-0.10.3/main/osx/ProxyWatcher.cpp <==
/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ProxyWatcher.h"
#include "NetworkPrefs.h"
#include "CFUtils.h"
#include "StringLess.h"
#include "GlobalState.h"
#include "ProxyDescriptor.h"
#include "InetAddr.h"
#include "AssertException.h"
#include <string>
#include <list>
#include <vector>

using namespace std;

class ProxyWatcher::GenericError
{
public:
	GenericError(std::string const& msg);
	
	GenericError(char const* msg);
	
	~GenericError();
	
	std::string const& getMessage() const { return m_msg; }
private:
	std::string m_msg;
};
	

ProxyWatcher::ProxyWatcher(
	StringRef const& client_name, RefPtr<CFRunLoopRef> const& run_loop)
try
:	m_ptrClientName(client_name),
	m_ptrRunLoop(run_loop),
	m_ptrNetworkStateKey(
		SCDynamicStoreKeyCreateNetworkGlobalEntity(
			NULL, kSCDynamicStoreDomainState, kSCEntNetIPv4
		),
		NORETAIN
	),
	m_ptrProxiesStateKey(
		SCDynamicStoreKeyCreateNetworkGlobalEntity(
			NULL, kSCDynamicStoreDomainState, kSCEntNetProxies
		),
		NORETAIN
	),
	m_bfPrefix(CFSTR("(BF) "))
{
	SCDynamicStoreContext context;
	context.version = 0;
	context.info = this;
	context.retain = 0;
	context.release = 0;
	context.copyDescription = 0;
	
	m_ptrStore = RefPtr<SCDynamicStoreRef>(
		SCDynamicStoreCreate(NULL, m_ptrClientName, &storeModifiedCallback, &context),
		NORETAIN
	);
	if (!m_ptrStore) {
		throw GenericError("SCDynamicStoreCreate() failed");
	}
	
	m_ptrRunLoopSource = RefPtr<CFRunLoopSourceRef>(
		SCDynamicStoreCreateRunLoopSource(NULL, m_ptrStore, 0), NORETAIN
	);
	if (!m_ptrRunLoopSource) {
		throw GenericError("SCDynamicStoreCreateRunLoopSource() failed");
	}
	
	CFRunLoopAddSource(m_ptrRunLoop, m_ptrRunLoopSource, kCFRunLoopCommonModes);
	
	ArrayRef keys = CFUtils::createCFTypeArray(
		m_ptrNetworkStateKey.get(), m_ptrProxiesStateKey.get(), 0
	);
	if (!SCDynamicStoreSetNotificationKeys(m_ptrStore, keys, NULL)) {
		throw GenericError("SCDynamicStoreSetNotificationKeys() failed");
	} 
		
	updateProxies();
} catch (GenericError const& e) {
	setGenericError(e.getMessage());
} catch (AssertException const& e) {
	setGenericError(e.getMessage());
} catch (std::exception const& e) {
	setGenericError(string("Exception: ")+e.what());
}

ProxyWatcher::~ProxyWatcher()
{
	if (m_ptrRunLoopSource) {
		CFRunLoopRemoveSource(m_ptrRunLoop, m_ptrRunLoopSource, kCFRunLoopCommonModes);
	}
}

void
ProxyWatcher::updateProxies()
try {
	DictRef network_state(
		(CFDictionaryRef)SCDynamicStoreCopyValue(m_ptrStore, m_ptrNetworkStateKey),
		NORETAIN
	);
	if (!network_state) {
		/*
		This means no network connection is available.  It can happen
		during location changes.  In this case, we'll receive another
		notification later.
		We don't reset proxy settings here, because it makes no difference
		when the network is down.
		*/
		return;
	}
	
	StringRef primary_interface(
		(CFStringRef)CFDictionaryGetValue(
			network_state, kSCDynamicStorePropNetPrimaryInterface
		),
		RETAIN
	);
	if (!primary_interface || CFGetTypeID(primary_interface) != CFStringGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	StringRef primary_service(
		(CFStringRef)CFDictionaryGetValue(
			network_state, kSCDynamicStorePropNetPrimaryService
		),
		RETAIN
	);
	if (!primary_service || CFGetTypeID(primary_service) != CFStringGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	StringRef service_path = CFUtils::composePath(kSCPrefNetworkServices, primary_service.get(), 0);
	
	getProxiesFromPersistentStore(primary_interface, service_path);
} catch (GenericError const& e) {
	setGenericError(e.getMessage());
} catch (AssertException const& e) {
	setGenericError(e.getMessage());
} catch (std::exception const& e) {
	setGenericError(string("Exception: ")+e.what());
}

void
ProxyWatcher::storeModifiedCallback(
	SCDynamicStoreRef store, CFArrayRef changedKeys, void* info)
{
	ProxyWatcher* obj = static_cast<ProxyWatcher*>(info);
	obj->updateProxies();
}

void
ProxyWatcher::getProxiesFromPersistentStore(
	StringRef const& primary_interface, StringRef const& primary_service)
{	
	NetworkPrefs prefs(m_ptrClientName);
	
	StringPair cur_set = findSetReferencingService(prefs, primary_service);
	if (!cur_set.first) {
		cur_set = getCurrentSet(prefs);
	}
	StringRef const& set_name = cur_set.second;
	
	StringRef service_path;
	if (!CFStringHasPrefix(set_name, m_bfPrefix)) {
		service_path = primary_service;
	} else { // has (BF) prefix
		CFIndex name_len = CFStringGetLength(set_name);
		CFIndex bf_prefix_len = CFStringGetLength(m_bfPrefix);
		CFRange range = CFRangeMake(bf_prefix_len, name_len - bf_prefix_len);
		StringRef counterpart_set_name(
			CFStringCreateWithSubstring(NULL, set_name, range),
			NORETAIN
		);
		StringRef counterpart_set_id = findSetByName(prefs, counterpart_set_name);
		if (!counterpart_set_id) {
			throw GenericError(
				"The current (BF) Location doesn't have a non-(BF) counterpart"
			);
		}
		
		service_path = findServiceBySetAndInterface(
			prefs, counterpart_set_id, primary_interface
		);
		if (!service_path) {
			throw GenericError(
				"(BF) Locations are out-of-sync with their counterparts.\n"
				"Reinstalling BFilter will fix that."
			);
		}
	}
	
	DictRef service(
		SCPreferencesPathGetValue(prefs.getRep(), service_path),
		RETAIN
	);
	if (!service || CFGetTypeID(service) != CFDictionaryGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	getProxiesFromService(service);
}

ProxyWatcher::StringPair
ProxyWatcher::findSetReferencingService(
	NetworkPrefs& prefs, StringRef const& service_path)
{
	typedef vector<StringPair> Sets;
	Sets sets;
	DECLARE_INSERTER(SetInserter, StringPair, Sets, c.push_back(val));
	SetInserter set_inserter(sets);
	prefs.enumSetsReferencingService(service_path, set_inserter);
	
	if (sets.size() == 1) {
		return sets[0];
	} else {
		return std::make_pair(StringRef(), StringRef());
	}
}

ProxyWatcher::StringPair
ProxyWatcher::getCurrentSet(NetworkPrefs& prefs)
{
	StringRef cur_set_path(
		(CFStringRef)SCPreferencesGetValue(prefs.getRep(), kSCPrefCurrentSet),
		RETAIN
	);
	if (!cur_set_path || CFGetTypeID(cur_set_path) != CFStringGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	DictRef set(
		SCPreferencesPathGetValue(prefs.getRep(), cur_set_path),
		RETAIN
	);
	if (!set) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	StringRef name(
		(CFStringRef)CFDictionaryGetValue(set, kSCPropUserDefinedName),
		RETAIN
	);
	if (!name || CFGetTypeID(name) != CFStringGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	StringRef set_id = CFUtils::extractLastPathComponent(cur_set_path);
	return std::make_pair(set_id, name);
}

StringRef
ProxyWatcher::findSetByName(NetworkPrefs& prefs, StringRef const& name)
{
	class SetFinder : public OutputFunction<StringPair>
	{
	public:
		SetFinder(StringRef const& set_name)
		: m_setName(set_name), m_setId() {}
		
		virtual void operator()(StringPair const& set) {
			if (!m_setId && CFEqual(set.second, m_setName)) {
				m_setId = set.first;
			}
		}
		
		StringRef const& getSetId() const { return m_setId; }
	private:
		StringRef m_setName;
		StringRef m_setId;
	};
	
	SetFinder set_finder(name);
	prefs.enumSets(set_finder);
	return set_finder.getSetId();
}

StringRef
ProxyWatcher::findServiceBySetAndInterface(
	NetworkPrefs& prefs, StringRef const& set_id, StringRef const& interface)
{
	typedef vector<StringRef> Services;
	Services services;
	DECLARE_INSERTER(ServiceInserter, StringPair, Services, c.push_back(val.second));
	ServiceInserter service_inserter(services);
	prefs.enumServicesReferencedBySet(set_id, service_inserter);
	
	Services::iterator it = services.begin();
	Services::iterator const end = services.end();
	for (; it != end; ++it) {
		StringRef path = CFUtils::composePath(*it, kSCEntNetInterface, 0);
		DictRef iface_dict(
			SCPreferencesPathGetValue(prefs.getRep(), path),
			RETAIN
		);
		if (!iface_dict) {
			continue;
		}
		if (CFGetTypeID(iface_dict) != CFDictionaryGetTypeID()) {
			throw AssertException(__FILE__, __LINE__);
		}
		
		CFStringRef device_name = (CFStringRef)CFDictionaryGetValue(
			iface_dict, kSCPropNetInterfaceDeviceName
		);
		if (device_name && CFEqual(device_name, interface)) {
			return *it;
		}
	}
	
	return StringRef();
}

void
ProxyWatcher::getProxiesFromService(DictRef const& service)
{
	DictRef proxies(
		(CFDictionaryRef)CFDictionaryGetValue(service, kSCEntNetProxies),
		RETAIN
	);
	if (!proxies || CFGetTypeID(service) != CFDictionaryGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	if (isProxyEnabled(proxies,  kSCPropNetProxiesHTTPEnable)) {
		processHttpProxy(proxies);
	} else if (isProxyEnabled(proxies, kSCPropNetProxiesSOCKSEnable)) {
		processSocksProxy(proxies);
	} else if (isProxyEnabled(proxies, CFSTR("ProxyAutoDiscoveryEnable"))) {
		processProxyAutoDiscovery(proxies);
	} else if (isProxyEnabled(proxies, CFSTR("ProxyAutoConfigEnable"))) {
		processProxyAutoConfig(proxies);
	} else {
		setNoProxy();
	}
}

bool
ProxyWatcher::isProxyEnabled(DictRef const& proxies, CFStringRef key)
{
	CFNumberRef val = (CFNumberRef)CFDictionaryGetValue(proxies, key);
	if (!val || CFGetTypeID(val) != CFNumberGetTypeID()) {
		return false;
	}
	int enabled = 0;
	CFNumberGetValue(val, kCFNumberIntType, &enabled);
	return enabled != 0;
}

void
ProxyWatcher::processHttpProxy(DictRef const& proxies)
{
	CFStringRef host_str = (CFStringRef)CFDictionaryGetValue(
		proxies, kSCPropNetProxiesHTTPProxy
	);
	if (!host_str || CFGetTypeID(host_str) != CFStringGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	CFNumberRef port_num = (CFNumberRef)CFDictionaryGetValue(
		proxies, kSCPropNetProxiesHTTPPort
	);
	if (!port_num || CFGetTypeID(port_num) != CFNumberGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	char host[256];
	if (!CFStringGetCString(host_str, host, sizeof(host), kCFStringEncodingASCII)) {
		throw GenericError("Proxy host contains international characters or is too long");
	}
	
	int port = 0;
	CFNumberGetValue(port_num, kCFNumberIntType, &port);
	
	ProxyDescriptor proxy_desc(InetAddr(host, port));
	
	typedef std::list<std::string> SkipList;
	SkipList skip_list;
	DECLARE_INSERTER(SkipListInserter, std::string, SkipList, c.push_back(val));
	SkipListInserter skip_list_inserter(skip_list);
	readProxySkipList(proxies, skip_list_inserter);
	
	GlobalState::WriteAccessor global_state;
	global_state->config().setNextHopProxy(proxy_desc);
	global_state->config().setNextHopProxyEnabled(true);
	global_state->config().setProxySkipList(skip_list);
}

void
ProxyWatcher::processSocksProxy(DictRef const& proxies)
{
	setGenericError("Unsupported feature: SOCKS proxy");
}

void
ProxyWatcher::processProxyAutoDiscovery(DictRef const& proxies)
{
	setGenericError("Unsupported feature: proxy auto discovery");
}

void
ProxyWatcher::processProxyAutoConfig(DictRef const& proxies)
{
	setGenericError("Unsupported feature: proxy auto configuration");
}

void
ProxyWatcher::readProxySkipList(
	DictRef const& proxies, OutputFunction<std::string>& output)
{
	CFArrayRef arr = (CFArrayRef)CFDictionaryGetValue(
		proxies, kSCPropNetProxiesExceptionsList
	);
	if (!arr) {
		return;
	}
	if (CFGetTypeID(arr) != CFArrayGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	char item_str[256];
	
	CFIndex const count = CFArrayGetCount(arr);
	for (CFIndex i = 0; i < count; ++i) {
		CFStringRef item = (CFStringRef)CFArrayGetValueAtIndex(arr, i);
		if (!item || CFGetTypeID(item) != CFStringGetTypeID()) {
			throw AssertException(__FILE__, __LINE__);
		}
		
		if (!CFStringGetCString(item, item_str, sizeof(item_str), kCFStringEncodingASCII)) {
			// To support international domain names, we would have to convert
			// strings to Punicode, which Core Foundation doesn't support.
			continue;
		}
		
		output(string(item_str));
		output("."+string(item_str));
	}
	
	// TODO: support ExcludeSimpleDomains property.
	// This will require modifying the Config class.
}

void
ProxyWatcher::setNoProxy()
{
	ProxyDescriptor proxy(InetAddr("", -1));
	GlobalState::WriteAccessor global_state;
	global_state->config().setNextHopProxy(proxy);
	global_state->config().setNextHopProxyEnabled(false);
}

void
ProxyWatcher::setGenericError(std::string const& msg)
{
	ProxyDescriptor proxy(msg);
	GlobalState::WriteAccessor global_state;
	global_state->config().setNextHopProxy(proxy);
	global_state->config().setNextHopProxyEnabled(false);
}


/*================== ProxyWatcher::GenericError ====================*/

ProxyWatcher::GenericError::GenericError(std::string const& msg)
:	m_msg(msg)
{
}

ProxyWatcher::GenericError::GenericError(char const* msg)
:	m_msg(msg)
{
}

ProxyWatcher::GenericError::~GenericError()
{
}

==> bfilter_1.0.2.orig.tar.gz: bfilter-1.0.2/main/osx/ProxyWatcher.cpp <==
/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ProxyWatcher.h"
#include "NetworkPrefs.h"
#include "CFUtils.h"
#include "StringLess.h"
#include "GlobalState.h"
#include "ProxyDescriptor.h"
#include "SymbolicInetAddr.h"
#include "AssertException.h"
#include <string>
#include <list>
#include <vector>

using namespace std;

class ProxyWatcher::GenericError
{
public:
	GenericError(std::string const& msg);
	
	GenericError(char const* msg);
	
	~GenericError();
	
	std::string const& getMessage() const { return m_msg; }
private:
	std::string m_msg;
};
	

ProxyWatcher::ProxyWatcher(
	StringRef const& client_name, RefPtr<CFRunLoopRef> const& run_loop)
try
:	m_ptrClientName(client_name),
	m_ptrRunLoop(run_loop),
	m_ptrNetworkStateKey(
		SCDynamicStoreKeyCreateNetworkGlobalEntity(
			NULL, kSCDynamicStoreDomainState, kSCEntNetIPv4
		),
		NORETAIN
	),
	m_ptrProxiesStateKey(
		SCDynamicStoreKeyCreateNetworkGlobalEntity(
			NULL, kSCDynamicStoreDomainState, kSCEntNetProxies
		),
		NORETAIN
	),
	m_bfPrefix(CFSTR("(BF) "))
{
	SCDynamicStoreContext context;
	context.version = 0;
	context.info = this;
	context.retain = 0;
	context.release = 0;
	context.copyDescription = 0;
	
	m_ptrStore = RefPtr<SCDynamicStoreRef>(
		SCDynamicStoreCreate(NULL, m_ptrClientName, &storeModifiedCallback, &context),
		NORETAIN
	);
	if (!m_ptrStore) {
		throw GenericError("SCDynamicStoreCreate() failed");
	}
	
	m_ptrRunLoopSource = RefPtr<CFRunLoopSourceRef>(
		SCDynamicStoreCreateRunLoopSource(NULL, m_ptrStore, 0), NORETAIN
	);
	if (!m_ptrRunLoopSource) {
		throw GenericError("SCDynamicStoreCreateRunLoopSource() failed");
	}
	
	CFRunLoopAddSource(m_ptrRunLoop, m_ptrRunLoopSource, kCFRunLoopCommonModes);
	
	ArrayRef keys = CFUtils::createCFTypeArray(
		m_ptrNetworkStateKey.get(), m_ptrProxiesStateKey.get(), 0
	);
	if (!SCDynamicStoreSetNotificationKeys(m_ptrStore, keys, NULL)) {
		throw GenericError("SCDynamicStoreSetNotificationKeys() failed");
	} 
		
	updateProxies();
} catch (GenericError const& e) {
	setGenericError(e.getMessage());
} catch (AssertException const& e) {
	setGenericError(e.getMessage());
} catch (std::exception const& e) {
	setGenericError(string("Exception: ")+e.what());
}

ProxyWatcher::~ProxyWatcher()
{
	if (m_ptrRunLoopSource) {
		CFRunLoopRemoveSource(m_ptrRunLoop, m_ptrRunLoopSource, kCFRunLoopCommonModes);
	}
}

void
ProxyWatcher::updateProxies()
try {
	DictRef network_state(
		(CFDictionaryRef)SCDynamicStoreCopyValue(m_ptrStore, m_ptrNetworkStateKey),
		NORETAIN
	);
	if (!network_state) {
		/*
		This means no network connection is available.  It can happen
		during location changes.  In this case, we'll receive another
		notification later.
		We don't reset proxy settings here, because it makes no difference
		when the network is down.
		*/
		return;
	}
	
	StringRef primary_interface(
		(CFStringRef)CFDictionaryGetValue(
			network_state, kSCDynamicStorePropNetPrimaryInterface
		),
		RETAIN
	);
	if (!primary_interface || CFGetTypeID(primary_interface) != CFStringGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	StringRef primary_service(
		(CFStringRef)CFDictionaryGetValue(
			network_state, kSCDynamicStorePropNetPrimaryService
		),
		RETAIN
	);
	if (!primary_service || CFGetTypeID(primary_service) != CFStringGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	StringRef service_path = CFUtils::composePath(kSCPrefNetworkServices, primary_service.get(), 0);
	
	getProxiesFromPersistentStore(primary_interface, service_path);
} catch (GenericError const& e) {
	setGenericError(e.getMessage());
} catch (AssertException const& e) {
	setGenericError(e.getMessage());
} catch (std::exception const& e) {
	setGenericError(string("Exception: ")+e.what());
}

void
ProxyWatcher::storeModifiedCallback(
	SCDynamicStoreRef store, CFArrayRef changedKeys, void* info)
{
	ProxyWatcher* obj = static_cast<ProxyWatcher*>(info);
	obj->updateProxies();
}

void
ProxyWatcher::getProxiesFromPersistentStore(
	StringRef const& primary_interface, StringRef const& primary_service)
{	
	NetworkPrefs prefs(m_ptrClientName);
	
	StringPair cur_set = findSetReferencingService(prefs, primary_service);
	if (!cur_set.first) {
		cur_set = getCurrentSet(prefs);
	}
	StringRef const& set_name = cur_set.second;
	
	StringRef service_path;
	if (!CFStringHasPrefix(set_name, m_bfPrefix)) {
		service_path = primary_service;
	} else { // has (BF) prefix
		CFIndex name_len = CFStringGetLength(set_name);
		CFIndex bf_prefix_len = CFStringGetLength(m_bfPrefix);
		CFRange range = CFRangeMake(bf_prefix_len, name_len - bf_prefix_len);
		StringRef counterpart_set_name(
			CFStringCreateWithSubstring(NULL, set_name, range),
			NORETAIN
		);
		StringRef counterpart_set_id = findSetByName(prefs, counterpart_set_name);
		if (!counterpart_set_id) {
			throw GenericError(
				"The current (BF) Location doesn't have a non-(BF) counterpart"
			);
		}
		
		service_path = findServiceBySetAndInterface(
			prefs, counterpart_set_id, primary_interface
		);
		if (!service_path) {
			throw GenericError(
				"(BF) Locations are out-of-sync with their counterparts.\n"
				"Reinstalling BFilter will fix that."
			);
		}
	}
	
	DictRef service(
		SCPreferencesPathGetValue(prefs.getRep(), service_path),
		RETAIN
	);
	if (!service || CFGetTypeID(service) != CFDictionaryGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	getProxiesFromService(service);
}

ProxyWatcher::StringPair
ProxyWatcher::findSetReferencingService(
	NetworkPrefs& prefs, StringRef const& service_path)
{
	typedef vector<StringPair> Sets;
	Sets sets;
	DECLARE_INSERTER(SetInserter, StringPair, Sets, c.push_back(val));
	SetInserter set_inserter(sets);
	prefs.enumSetsReferencingService(service_path, set_inserter);
	
	if (sets.size() == 1) {
		return sets[0];
	} else {
		return std::make_pair(StringRef(), StringRef());
	}
}

ProxyWatcher::StringPair
ProxyWatcher::getCurrentSet(NetworkPrefs& prefs)
{
	StringRef cur_set_path(
		(CFStringRef)SCPreferencesGetValue(prefs.getRep(), kSCPrefCurrentSet),
		RETAIN
	);
	if (!cur_set_path || CFGetTypeID(cur_set_path) != CFStringGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	DictRef set(
		SCPreferencesPathGetValue(prefs.getRep(), cur_set_path),
		RETAIN
	);
	if (!set) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	StringRef name(
		(CFStringRef)CFDictionaryGetValue(set, kSCPropUserDefinedName),
		RETAIN
	);
	if (!name || CFGetTypeID(name) != CFStringGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	StringRef set_id = CFUtils::extractLastPathComponent(cur_set_path);
	return std::make_pair(set_id, name);
}

StringRef
ProxyWatcher::findSetByName(NetworkPrefs& prefs, StringRef const& name)
{
	class SetFinder : public OutputFunction<StringPair>
	{
	public:
		SetFinder(StringRef const& set_name)
		: m_setName(set_name), m_setId() {}
		
		virtual void operator()(StringPair const& set) {
			if (!m_setId && CFEqual(set.second, m_setName)) {
				m_setId = set.first;
			}
		}
		
		StringRef const& getSetId() const { return m_setId; }
	private:
		StringRef m_setName;
		StringRef m_setId;
	};
	
	SetFinder set_finder(name);
	prefs.enumSets(set_finder);
	return set_finder.getSetId();
}

StringRef
ProxyWatcher::findServiceBySetAndInterface(
	NetworkPrefs& prefs, StringRef const& set_id, StringRef const& interface)
{
	typedef vector<StringRef> Services;
	Services services;
	DECLARE_INSERTER(ServiceInserter, StringPair, Services, c.push_back(val.second));
	ServiceInserter service_inserter(services);
	prefs.enumServicesReferencedBySet(set_id, service_inserter);
	
	Services::iterator it = services.begin();
	Services::iterator const end = services.end();
	for (; it != end; ++it) {
		StringRef path = CFUtils::composePath(*it, kSCEntNetInterface, 0);
		DictRef iface_dict(
			SCPreferencesPathGetValue(prefs.getRep(), path),
			RETAIN
		);
		if (!iface_dict) {
			continue;
		}
		if (CFGetTypeID(iface_dict) != CFDictionaryGetTypeID()) {
			throw AssertException(__FILE__, __LINE__);
		}
		
		CFStringRef device_name = (CFStringRef)CFDictionaryGetValue(
			iface_dict, kSCPropNetInterfaceDeviceName
		);
		if (device_name && CFEqual(device_name, interface)) {
			return *it;
		}
	}
	
	return StringRef();
}

void
ProxyWatcher::getProxiesFromService(DictRef const& service)
{
	DictRef proxies(
		(CFDictionaryRef)CFDictionaryGetValue(service, kSCEntNetProxies),
		RETAIN
	);
	if (!proxies || CFGetTypeID(service) != CFDictionaryGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	if (isProxyEnabled(proxies,  kSCPropNetProxiesHTTPEnable)) {
		processHttpProxy(proxies);
	} else if (isProxyEnabled(proxies, kSCPropNetProxiesSOCKSEnable)) {
		processSocksProxy(proxies);
	} else if (isProxyEnabled(proxies, CFSTR("ProxyAutoDiscoveryEnable"))) {
		processProxyAutoDiscovery(proxies);
	} else if (isProxyEnabled(proxies, CFSTR("ProxyAutoConfigEnable"))) {
		processProxyAutoConfig(proxies);
	} else {
		setNoProxy();
	}
}

bool
ProxyWatcher::isProxyEnabled(DictRef const& proxies, CFStringRef key)
{
	CFNumberRef val = (CFNumberRef)CFDictionaryGetValue(proxies, key);
	if (!val || CFGetTypeID(val) != CFNumberGetTypeID()) {
		return false;
	}
	int enabled = 0;
	CFNumberGetValue(val, kCFNumberIntType, &enabled);
	return enabled != 0;
}

void
ProxyWatcher::processHttpProxy(DictRef const& proxies)
{
	CFStringRef host_str = (CFStringRef)CFDictionaryGetValue(
		proxies, kSCPropNetProxiesHTTPProxy
	);
	if (!host_str || CFGetTypeID(host_str) != CFStringGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	CFNumberRef port_num = (CFNumberRef)CFDictionaryGetValue(
		proxies, kSCPropNetProxiesHTTPPort
	);
	if (!port_num || CFGetTypeID(port_num) != CFNumberGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	char host[256];
	if (!CFStringGetCString(host_str, host, sizeof(host), kCFStringEncodingASCII)) {
		throw GenericError("Proxy host contains international characters or is too long");
	}
	
	int port = 0;
	CFNumberGetValue(port_num, kCFNumberIntType, &port);
	
	ProxyDescriptor proxy_desc(SymbolicInetAddr(host, port));
	
	typedef std::list<std::string> SkipList;
	SkipList skip_list;
	DECLARE_INSERTER(SkipListInserter, std::string, SkipList, c.push_back(val));
	SkipListInserter skip_list_inserter(skip_list);
	readProxySkipList(proxies, skip_list_inserter);
	
	GlobalState::WriteAccessor global_state;
	global_state->config().setNextHopProxy(proxy_desc);
	global_state->config().setNextHopProxyEnabled(true);
	global_state->config().setProxySkipList(skip_list);
}

void
ProxyWatcher::processSocksProxy(DictRef const& proxies)
{
	setGenericError("Unsupported feature: SOCKS proxy");
}

void
ProxyWatcher::processProxyAutoDiscovery(DictRef const& proxies)
{
	setGenericError("Unsupported feature: proxy auto discovery");
}

void
ProxyWatcher::processProxyAutoConfig(DictRef const& proxies)
{
	setGenericError("Unsupported feature: proxy auto configuration");
}

void
ProxyWatcher::readProxySkipList(
	DictRef const& proxies, OutputFunction<std::string>& output)
{
	CFArrayRef arr = (CFArrayRef)CFDictionaryGetValue(
		proxies, kSCPropNetProxiesExceptionsList
	);
	if (!arr) {
		return;
	}
	if (CFGetTypeID(arr) != CFArrayGetTypeID()) {
		throw AssertException(__FILE__, __LINE__);
	}
	
	char item_str[256];
	
	CFIndex const count = CFArrayGetCount(arr);
	for (CFIndex i = 0; i < count; ++i) {
		CFStringRef item = (CFStringRef)CFArrayGetValueAtIndex(arr, i);
		if (!item || CFGetTypeID(item) != CFStringGetTypeID()) {
			throw AssertException(__FILE__, __LINE__);
		}
		
		if (!CFStringGetCString(item, item_str, sizeof(item_str), kCFStringEncodingASCII)) {
			// To support international domain names, we would have to convert
			// strings to Punicode, which Core Foundation doesn't support.
			continue;
		}
		
		output(string(item_str));
		output("."+string(item_str));
	}
	
	// TODO: support ExcludeSimpleDomains property.
	// This will require modifying the Config class.
}

void
ProxyWatcher::setNoProxy()
{
	ProxyDescriptor proxy(SymbolicInetAddr("", -1));
	GlobalState::WriteAccessor global_state;
	global_state->config().setNextHopProxy(proxy);
	global_state->config().setNextHopProxyEnabled(false);
}

void
ProxyWatcher::setGenericError(std::string const& msg)
{
	ProxyDescriptor proxy(msg);
	GlobalState::WriteAccessor global_state;
	global_state->config().setNextHopProxy(proxy);
	global_state->config().setNextHopProxyEnabled(false);
}


/*================== ProxyWatcher::GenericError ====================*/

ProxyWatcher::GenericError::GenericError(std::string const& msg)
:	m_msg(msg)
{
}

ProxyWatcher::GenericError::GenericError(char const* msg)
:	m_msg(msg)
{
}

ProxyWatcher::GenericError::~GenericError()
{
}
